# flake8: noqa: E501
from pathlib import Path
from typing import Dict, List

# Base directories
BASE_DIR: Path = Path(__file__).resolve().parent.parent
csv_file: Path = BASE_DIR / "Tables/original_tables/ct_ecmwf.rc"
excel_file: Path = BASE_DIR / "Tables/original_tables/ct_ecmwf.xlsx"
json_output_path: Path = BASE_DIR / "Tables/era5-cmor-tables/Tables"
search_dirs: List[Path] = [
    BASE_DIR / "Tables/source_tables/obs4MIPs-cmor-tables/Tables",
    BASE_DIR / "Tables/source_tables/cmip6-cmor-tables/Tables",
]

# Frequency priority
frequency_priority: Dict[str, List[str]] = {
    "1hr": [
        "A1hr",
        "A3hr",
        "A6hr",  # Atmospheric hourly tables
        "E1hr",
        "E3hr",
        "E6hrZ",  # Earth system / energy hourly
        "E1hrClimMon",
        "E3hrPt",  # Other energy system hourly
        "CF3hr",
        "CFsubhr",
        "Esubhr",  # High-res and subhourly
        "3hr",
        "6hrLev",
        "6hrPlev",
        "6hrPlevPt",  # More coarse-hourly or pressure-level
        "AERhr",  # Aerosol-specific hourly
    ],
    "day": [
        "Aday",
        "Eday",
        "EdayZ",  # Standard and energy/dynamics daily
        "CFday",  # High-res daily
        "SIday",  # Sea ice daily
        "Oday",  # Ocean daily
        "AERday",  # Aerosol daily
        "day",  # Generic fallback
    ],
    "mon": [
        "Amon",
        "Lmon",
        "Omon",  # Atmosphere, land, ocean monthly
        "Emon",
        "EmonZ",  # Energy system monthly
        "CFmon",  # High-res monthly
        "SImon",  # Sea ice monthly
        "AERmon",
        "AERmonZ",  # Aerosol monthly
        "ImonAnt",
        "ImonGre",
        "LImon",  # Ice-specific monthly
    ],
    "fx": [
        "fx",
        "Ofx",
        "Efx",
        "IfxAnt",
        "IfxGre",  # Fixed fields (land, ocean, earth system, ice)
    ],
}

# Level groupings
level_categories: Dict[str, List[str]] = {
    "sfc": ["sfc_an", "sfc_fc"],
    "sfc_land": ["sfc_an_land", "sfc_fc_land"],
    "pl": ["pl_an", "pl_fc"],
    "ml": ["ml_an"],
}

# Realm → prefix
realm_prefix_map: Dict[str, str] = {
    "aerosol": "AER",
    "atmos": "A",
    "atmosChem": "AER",
    "ice": "I",
    "land": "L",
    "landIce": "LI",
    "ocean": "O",
    "seaIce": "SI",
}

# Approximate intervals in days
approx_interval_map: Dict[str, float] = {
    "1hr": round(1 / 24, 5),  # 0.04167
    "day": 1.00000,
    "mon": 30.00000,
    "fx": 0.00000,
}

# Number of levels
level_number: Dict[str, int] = {"pl": 37, "ml": 137}

# Special variables not necessarily found as ECMWF variables or @ DKRZ
extra_variables: Dict[str, Dict] = {
    "areacella": {
        "frequencies": ["fx"],
        "levels": ["sfc", "sfc_land"],
        "level_types": {"sfc": "sfc_an", "sfc_land": "sfc_an_land"},
        "grids": {"sfc": "redGG-N320", "sfc_land": "redGG-N1280"},
        "table": "fx",
        "mapping": "Obs4MIPs",
    },
    "hurs": {
        "frequencies": ["1hr", "day", "mon"],
        "levels": ["sfc", "sfc_land"],
        "level_types": {"sfc": "sfc_an", "sfc_land": "sfc_an_land"},
        "grids": {"sfc": "redGG-N320", "sfc_land": "redGG-N1280"},
        "comment": "This variable is NOT produced by ECMWF. It was derived from tdps and tas from the saturation vapor pressure fraction: e_sat_tdps/e_sat_tas where e_sat is: https://metview.readthedocs.io/en/latest/api/functions/saturation_vapour_pressure.html#saturation_vapour_pressure",
    },
    "sfcWind": {
        "frequencies": ["1hr", "day", "mon"],
        "levels": ["sfc", "sfc_land"],
        "level_types": {"sfc": "sfc_an", "sfc_land": "sfc_an_land"},
        "grids": {"sfc": "redGG-N320", "sfc_land": "redGG-N1280"},
        "comment": "This variable is NOT produced by ECMWF. It was derived from UAS and VAS components by means of sqrt[uas*uas + vas*vas]",
    },
}
