import argparse
from argparse import RawTextHelpFormatter
from pathlib import Path

from config import csv_file, excel_file, json_output_path, search_dirs
from era5_tables.converter import csv_to_cmor_json, harmonize_variables
from era5_tables.utils import csv_to_excel, excel_to_csv

table_sources = [p.parent.name for p in search_dirs]
common_root = str(search_dirs[0].parent.parent)


def main():
    parser = argparse.ArgumentParser(
        description=(
            "Convert CMOR table data between CSV, Excel, and JSON formats. "
            f"\nIt uses the CMOR JSON schema, based on {table_sources}, "
            f"located in {common_root}. (as submodules)"
        ),
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "--csv-to-json",
        action="store_true",
        help="Convert CSV to CMOR JSON files",
    )
    parser.add_argument(
        "--csv-to-excel", action="store_true", help="Convert CSV to Excel file"
    )
    parser.add_argument(
        "--excel-to-csv", action="store_true", help="Convert Excel file to CSV"
    )

    parser.add_argument(
        "--csv",
        type=str,
        default=csv_file,
        help=f"Path to input CSV file (default: {csv_file})",
    )
    parser.add_argument(
        "--excel",
        type=str,
        default=excel_file,
        help=f"Path to Excel file (default: {excel_file})",
    )
    parser.add_argument(
        "--json",
        type=str,
        default=json_output_path,
        help=f"Path to output file or directory (default: {json_output_path})",
    )

    parser.add_argument("--var", nargs="+", help="Variable(s) to include")
    parser.add_argument(
        "--freq", nargs="+", help="Frequency/frequencies to include"
    )
    parser.add_argument("--ltype", nargs="+", help="Level type(s) to include")

    parser.add_argument(
        "--clean",
        action="store_true",
        help="Clean output JSON dir before generation. (no need of --csv-to-json)",
    )

    parser.add_argument(
        "-ee",
        "--exclude-extra",
        action="store_true",
        default=False,
        help="Exclude extra variables from processing",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Print each variable being processed",
    )

    args = parser.parse_args()

    if args.csv_to_json:
        if not args.csv or not args.json:
            parser.error("--csv and --json is required for --csv-to-json")
        csv_path = Path(args.csv)
        json_path = Path(args.json)
        csv_to_cmor_json(
            csv_filepath=csv_path,
            json_output_path=json_path,
            var=args.var,
            freq=args.freq,
            ltype=args.ltype,
            exclude_extra_variables=args.exclude_extra,
            verbose=args.verbose,
        )
        harmonize_variables(json_dir=json_path, verbose=args.verbose)

    elif args.clean:
        if not args.json:
            parser.error("--json is required for --clean")
        json_path = Path(args.json)
        csv_to_cmor_json(
            json_output_path=json_path,
            clean_output=True,
        )

    elif args.csv_to_excel:
        if not args.csv or not args.excel:
            parser.error("--csv and --excel are required for --csv-to-excel")
        csv_to_excel(args.csv, args.excel)

    elif args.excel_to_csv:
        if not args.excel or not args.csv:
            parser.error("--excel and --csv are required for --excel-to-csv")
        excel_to_csv(args.excel, args.csv)

    else:
        parser.print_help()


if __name__ == "__main__":
    main()
