import glob
import json
import os
import re
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import Any, Dict, Iterable, List, Literal, Optional, Union

import pandas as pd
from tqdm import tqdm

from config import (
    approx_interval_map,
    csv_file,
    extra_variables,
    frequency_priority,
    json_output_path,
    level_categories,
    level_number,
    realm_prefix_map,
    search_dirs,
)

GridType = Literal["redGG-N1280", "redGG-N320", "specG-T639", ""]
today_date: str = datetime.today().strftime("%d %B %Y")


def _determine_level_category(level_type: str) -> str:
    """
    Determines the level category (e.g., 'sfc', 'pl', etc.) for a given level
    type.

    Parameters:
        level_type (str): The level type string from the dataset.

    Returns:
        str: The corresponding level category.
    """
    for category, values in level_categories.items():
        if level_type in values:
            return category
    return "sfc"


def _load_variable_from_table(
    table_path: str, variable_name: str, table_prefix: str
) -> Optional[Dict]:
    """
    Loads a variable definition from a JSON table file.

    Parameters:
        table_path (str): Path to the JSON file.
        variable_name (str): Name of the variable to load.
        table_prefix (str): Prefix used to locate nested entries in the file.

    Returns:
        Optional[Dict]: The variable dictionary if found, else None.
    """

    try:
        with open(table_path, "r") as f:
            data = json.load(f)

        if "variable_entry" in data:
            return data["variable_entry"].get(variable_name)
        if table_prefix in data and "variable_entry" in data[table_prefix]:
            return data[table_prefix]["variable_entry"].get(variable_name)

        return None
    except Exception:
        return None


def _find_best_matching_variable(
    variable: str, frequency: str, search_dirs: List[Path]
) -> Optional[Dict]:
    """
    Searches through priority CMOR tables to find the best match for a variable.

    Parameters:
        variable (str): Name of the variable to find.
        frequency (str): Target frequency (e.g., '1hr', 'day').
        search_dirs (List[str]): List of directories to search.

    Returns:
        Optional[Dict]: Matching variable metadata if found.
    """

    priority_tables: List[str] = frequency_priority.get(frequency, [])
    for search_dir in search_dirs:
        for table_prefix in priority_tables:
            for root, _, files in os.walk(search_dir):
                matching_files: List[str] = [
                    file
                    for file in files
                    if (
                        file.endswith(".json")
                        and f"_{table_prefix}" in file  # noqa: W503
                        or file == f"{table_prefix}.json"
                    )
                ]
                for file in matching_files:
                    table_path: str = os.path.join(root, file)
                    variable_data: Dict | None = _load_variable_from_table(
                        table_path, variable, table_prefix
                    )
                    if variable_data:
                        variable_data["source_table"] = os.path.basename(file)
                        return variable_data
    return None


def _read_csv(csv_filepath: Path) -> pd.DataFrame:
    """
    Reads a custom-format CMOR configuration CSV.

    Parameters:
        csv_filepath (str): Path to the .rc file.

    Returns:
        pd.DataFrame: Cleaned DataFrame with column names and filled missing
            values.
    """
    with open(csv_filepath, "r", encoding="utf-8") as f:
        lines: List[str] = f.readlines()
    header_line_idx: int = next(
        idx for idx, line in enumerate(lines) if line.startswith("#CCC|")
    )
    headers: List[str] = lines[header_line_idx].strip().lstrip("#").split("|")
    data_lines: List[str] = lines[header_line_idx + 1 :]  # noqa: E203
    df: pd.DataFrame = pd.read_csv(
        StringIO("".join(data_lines)), sep="|", names=headers, engine="python"
    )
    df.columns = df.columns.str.strip()
    return df.dropna(subset=["CMPAR"]).fillna("")


def _get_mapping_source(cmip_val, source_table: Optional[str] = "") -> str:
    """
    Resolves the mapping origin for a variable (e.g., 'ECMWF', 'CMIP6').
    The mapping goes as follows in order of priority: Obs4MIPs, CMIP6, CF,
    ECMWF. When possible it looks at Obs4MIPs tables, then CMIP6, if not it
    relies on the CF CV that was manually written in the csv file. Finally,
    for new variables we resort to ECMWF naming.

    Parameters:
        cmip_val (int or str): Mapping indicator value.
        source_table (str, optional): Name of the source table file.

    Returns:
        str: The mapping label.
    """
    try:
        cmip_val = int(cmip_val)
    except (ValueError, TypeError):
        return "unknown"

    if cmip_val == 0:
        return "ECMWF"
    elif cmip_val == 1:
        return "CF"
    elif cmip_val == 6:
        lower_table: str = source_table.lower() if source_table else ""
        if "obs4mips" in lower_table:
            return "obs4MIPs"
        elif "cmip6" in lower_table:
            return "CMIP6"
        else:
            return "CMIP6"
    return "ECMWF"


def _filter_level_grid(row: dict, level_type: str) -> dict:
    """
    Extracts the correct orig_grid and level_type based on the specified level.

    Parameters:
        row (dict): Row from the CSV file as a dictionary.
        level_type (str): Target level_type (e.g., 'sfc_fc_land').

    Returns:
        dict: Dictionary with keys 'level_type' and 'orig_grid'.
    """
    all_level_types: list[str] = [
        lt.strip() for lt in row.get("LTYPE", "").split(",") if lt.strip()
    ]
    all_grids: list[str] = [
        g.strip() for g in row.get("ECGRID", "").split(",") if g.strip()
    ]

    if level_type not in all_level_types:
        return {"level_type": "", "orig_grid": ""}

    grid: GridType = ""

    if level_type.startswith("sfc"):
        grid = "redGG-N1280" if "land" in level_type else "redGG-N320"

    elif level_type.startswith("ml"):
        if "specG-T639" in all_grids:
            grid = "specG-T639"
        elif "redGG-N320" in all_grids:
            grid = "redGG-N320"

    elif level_type.startswith("pl"):
        grid = "redGG-N320"

    return {"level_type": level_type, "orig_grid": grid}


def _guess_realm_freq(realm: str, frequency: str) -> str:
    """
    Guesses the CMOR table name based on realm and frequency.

    Parameters:
        realm (str): Modeling realm (e.g., 'atmos', 'land').
        frequency (str): Frequency name ('1hr', 'day', etc.).

    Returns:
        str: Inferred table name (e.g., 'Aday').
    """

    for key, prefix in realm_prefix_map.items():
        if key in realm:
            return f"{prefix}{frequency}"
    return frequency


def _extract_table_name(source_table: str) -> str:
    """
    Extracts the short table name from a full filename.

    Parameters:
        source_table (str): Filename (e.g., 'CMIP6_Aday.json').

    Returns:
        str: Short table name (e.g., 'Aday').
    """

    return Path(source_table).stem.split("_")[-1] if source_table else ""


def _get_dimensions(
    matched: Optional[Dict], frequency: str, level_type: str
) -> str:
    """
    Build dimensions string, resolving appropriate vertical level and ensuring
    no duplicates. Replaces time-like variants (e.g. time1, time2) with
    standard 'time'.
    """
    default_dims: (
        Literal["longitude latitude"] | Literal["longitude latitude time"]
    ) = (
        "longitude latitude"
        if frequency == "fx"
        else "longitude latitude time"
    )
    dims: str = (
        matched.get("dimensions", default_dims) if matched else default_dims
    )
    dims_parts = dims.split()
    dims_parts = [
        "time" if re.match(r"time\d*$", dim) else dim for dim in dims_parts
    ]
    dims_parts = ["longitude" if dim == "site" else dim for dim in dims_parts]
    if "longitude" in dims_parts and "latitude" not in dims_parts:
        dims_parts.insert(dims_parts.index("longitude") + 1, "latitude")
    dims_parts = [
        dim
        for dim in dims_parts
        if not (dim.startswith("plev") or dim.startswith("alevel"))
    ]

    time_indices: List[int] = [
        i for i, dim in enumerate(dims_parts) if re.match(r"time\d*$", dim)
    ]
    time_idx: int = time_indices[0] if time_indices else len(dims_parts)
    if level_type.startswith("pl"):
        dims_parts.insert(time_idx, f"plev{level_number['pl']}")
    elif level_type.startswith("ml"):
        dims_parts.insert(time_idx, f"alevel{level_number['ml']}")

    return " ".join(dims_parts)


def _get_cell_info(
    var_name: str, matched: Optional[Dict], frequency: str, modeling_realm: str
) -> dict:
    """
    Determines `cell_methods` and `cell_measures`, preferring matched metadata
    when available.

    Parameters:
        var_name (str): Name of the variable.
        matched (Optional[Dict]): Matching table variable info.
        frequency (str): Frequency string.
        modeling_realm (str): Modeling realm (e.g., 'ocean', 'land').

    Returns:
        dict: Dictionary with 'cell_methods' and 'cell_measures'.
    """

    realm_lc: str = modeling_realm.lower()

    time_method = "time: mean"
    area_method = "area: mean"
    default_measure = "area: areacella"

    if frequency in {"1hr", "day", "mon"} and any(
        ext in var_name.lower() for ext in ["max", "min"]
    ):
        if "max" in var_name.lower():
            extreme = "maximum"
        elif "min" in var_name.lower():
            extreme = "minimum"
        if frequency == "mon":
            time_method = f"time: {extreme} within days time: mean over days"
        else:
            time_method = f"time: {extreme}"
    if "ocean" in realm_lc or "seaice" in realm_lc:
        area_method = "area: mean where sea"
        default_measure = "area: areacello"
    elif "land" in realm_lc:
        area_method = "area: mean where land"
    elif "landice" in realm_lc:
        area_method = "area: mean where snow"
    default_method: str = f"{area_method} {time_method}"

    matched_method: str = matched.get("cell_methods", "") if matched else ""
    matched_measure: str = matched.get("cell_measures", "") if matched else ""

    cell_methods = matched_method.strip() or default_method
    cell_measures = matched_measure.strip() or default_measure
    return {"cell_methods": cell_methods, "cell_measures": cell_measures}


def _build_cmor_entry(
    var_name: str,
    matched: Optional[Dict],
    row: pd.Series,
    frequency: str,
    level_type: str,
) -> Dict:
    """
    Builds a CMOR variable entry by combining CSV, table, and derived values.

    Parameters:
        var_name (str): Variable name.
        matched (Optional[Dict]): Matching variable from external CMOR tables.
        row (pd.Series): Row from the CSV config.
        frequency (str): Frequency string.
        level_type (str): Level type.

    Returns:
        Dict: Final CMOR variable entry.
    """
    comment = row.get("COMMENT", "") or ""
    long_name = row["CMLNAME"]
    realm = row["REALM"]
    standard_name = row["CFNAME"]
    units = row["CMUNIT"]

    level_grid = _filter_level_grid(row, level_type)
    source_table = matched.get("source_table", "") if matched else ""
    mapping: str = _get_mapping_source(row["CMIP"], source_table)
    modeling_realm = matched.get("modeling_realm", "") if matched else realm
    cell_info = _get_cell_info(var_name, matched, frequency, modeling_realm)
    matched_comment = matched.get("comment", "") if matched else ""
    final_comment = (
        f"{matched_comment}. {comment}".strip(". ")
        if comment
        else matched_comment
    )

    return {
        "cell_measures": cell_info["cell_measures"],
        "cell_methods": cell_info["cell_methods"],
        "comment": final_comment,
        "dimensions": _get_dimensions(matched, frequency, level_type),
        "frequency": frequency,
        "long_name": long_name,
        "modeling_realm": modeling_realm,
        "ok_max_mean_abs": (
            matched.get("ok_max_mean_abs", "") if matched else ""
        ),
        "ok_min_mean_abs": (
            matched.get("ok_min_mean_abs", "") if matched else ""
        ),
        "out_name": var_name,
        "positive": matched.get("positive", "") if matched else "",
        "standard_name": standard_name,
        "type": matched.get("type", "real") if matched else "real",
        "units": units,
        "valid_max": matched.get("valid_max", "") if matched else "",
        "valid_min": matched.get("valid_min", "") if matched else "",
        # Additional metadata:
        "grib_table": str(int(row["ECTABLE"])),
        "grib_paramID": str(int(row["ECCODE"])),
        "orig_short_name": row["ECPAR"],
        "orig_name": row["ECNAME"],
        "orig_units": row["ECUNIT"],
        "grib_description": row["ECDESC"],
        "orig_grid": level_grid["orig_grid"],
        "level_type": level_grid["level_type"],
        "conversion": row["CMFACT"],
        "source_table": matched.get("source_table", "") if matched else "",
        "table": (
            _extract_table_name(source_table)
            if source_table
            else _guess_realm_freq(realm, frequency)
        ),
        "mapping": mapping,
        "DKRZ_ID": row["CCC"],
    }


def _add_extra_variables(
    grouped_json: dict, search_dirs: List[Path], verbose: bool = False
) -> None:
    """
    Adds extra variables defined in config.extra_variables into grouped_json,
    merging matched info and filling in configured metadata.

    Parameters:
        grouped_json (dict): A dictionary containing the current grouped
            configuration of variables with their metadata.
        search_dirs (List[Path]): List of directory paths to search for matching
            variable metadata in external tables.
        verbose (bool, optional): Flag to control the verbosity of the function's
            output. If True, additional information is printed. Default is False.

    Returns:
        None: The function modifies the grouped_json in place.
    """
    iterable = extra_variables.items()
    if not verbose:
        iterable = tqdm(
            iterable,
            total=len(extra_variables.items()),
            desc="➕ Adding extra variables",
            colour="magenta",
        )
    for var_name, meta in iterable:
        frequencies = meta.get("frequencies", [])
        levels = meta.get("levels", [])
        configured_comment = meta.get("comment", "")
        for freq in frequencies:
            for level in levels:
                key = (freq, level)
                if var_name in grouped_json.get(key, {}):
                    continue

                matched = _find_best_matching_variable(
                    var_name, freq, search_dirs
                )
                if not matched:
                    print(
                        f"⚠️  Warning: '{var_name}' not found in source tables."
                    )
                    continue

                source_table = matched.pop("source_table", "")
                matched_comment = matched.get("comment", "")
                final_comment = (
                    f"{matched_comment} {configured_comment}".strip(". ")
                    if configured_comment
                    else matched_comment
                )

                # Override frequency with the one from config (not from matched)
                matched["frequency"] = freq
                matched["comment"] = final_comment

                table_name = meta.get("table") or _extract_table_name(
                    source_table
                )
                mapping = meta.get("mapping") or _get_mapping_source(
                    6, source_table
                )
                level_type = meta.get("level_types", {}).get(level, "")
                grid = meta.get("grids", {}).get(level, "")

                extra_metadata: dict[str, str] = {
                    "grib_table": "",
                    "grib_paramID": "",
                    "orig_short_name": "",
                    "orig_name": "",
                    "orig_units": "",
                    "grib_description": "",
                    "orig_grid": grid,
                    "level_type": level_type,
                    "conversion": "",
                    "source_table": source_table,
                    "table": table_name,
                    "mapping": mapping,
                    "DKRZ_ID": meta.get("CCC", ""),
                }

                grouped_json.setdefault(key, {})[var_name] = {
                    **matched,
                    **extra_metadata,
                }
                if verbose:
                    print(f"➕ Added '{var_name}' ➤ {key}")


def _str_to_list(val) -> Optional[List[str]]:
    """
    Ensures a value is a list; wraps a single string into a list.

    Parameters:
        val (str or list): Input value.

    Returns:
        Optional[List[str]]: List version or None.
    """
    if val is None:
        return None
    return val if isinstance(val, list) else [val]


def csv_to_cmor_json(
    csv_filepath: Path = csv_file,
    json_output_path: Path = json_output_path,
    var: Optional[Union[str, List[str]]] = None,
    freq: Optional[Union[str, List[str]]] = None,
    ltype: Optional[Union[str, List[str]]] = None,
    clean_output: bool = False,
    exclude_extra_variables: bool = False,
    verbose: bool = False,
) -> None:
    """
    Main function to convert .rc CSV variable definitions to CMOR JSON files.

    Parameters:
        csv_filepath (Path): Path to the input CSV file.
        json_output_path (Path): Output directory for the generated JSONs.
        var (Optional[str or List[str]]): Variable(s) to include.
        freq (Optional[str or List[str]]): Frequencies to include.
        ltype (Optional[str or List[str]]): Level types to include.
        clean_output (bool): Whether to clean output directory before writing.
        exclude_extra_variables (bool): Whether to exclude extra variables from
            the processing.
        verbose (bool): Whether to show per-variable output.

    Returns:
        None: The function creates JSON files in the output directory.
    """

    var = _str_to_list(var)
    freq = _str_to_list(freq)
    ltype = _str_to_list(ltype)

    if clean_output:
        files_to_clean = glob.glob(os.path.join(json_output_path, "*.json"))
        if len(files_to_clean) < 1:
            print(
                "⚠️  No JSON files found in output directory. Nothing to clean."
            )
            return
        files_to_clean = tqdm(
            files_to_clean,
            total=len(files_to_clean),
            desc=f"🧹 Removing JSON files from: {json_output_path}",
            colour="red",
        )
        for file in files_to_clean:
            os.remove(file)
        return

    df: pd.DataFrame = _read_csv(csv_filepath)
    if var:
        df = df[df["CMPAR"].isin(var)]

    grouped_json: dict[tuple[str, str], dict[str, dict]] = {}

    iterable: Iterable[tuple[int, pd.Series]] = df.iterrows()
    if not verbose:
        iterable = tqdm(
            iterable,
            total=len(df),
            desc="🔄 Processing variables",
            colour="blue",
        )

    for _, row in iterable:
        var_name: str = row["CMPAR"]
        time_representation: list[str] = [
            t.strip().upper()
            for t in row["TREPR"].split(",")
            if pd.notna(row["TREPR"])
        ]
        level_types: list[str] = [
            lt.strip()
            for lt in row["LTYPE"].split(",")
            if pd.notna(row["LTYPE"])
        ]

        for level_type in level_types:
            level_group = _determine_level_category(level_type)
            if ltype and level_group not in ltype:
                continue

            if set(time_representation) == {"INV"}:
                applicable_frequencies: List[str] = ["fx"]
            else:
                applicable_frequencies = ["1hr", "day", "mon"]
                if "INV" in time_representation:
                    applicable_frequencies.append("fx")

            for frequency in applicable_frequencies:
                if freq and frequency not in freq:
                    continue
                if verbose:
                    print(
                        f"🔄 Processing variable: {var_name:15} | {frequency:4} | "
                        f"Level: {level_group}"
                    )
                key: tuple[str, str] = (frequency, level_group)
                matched: dict[Any, Any] | None = _find_best_matching_variable(
                    var_name, frequency, search_dirs
                )
                cmor_entry: dict = _build_cmor_entry(
                    var_name, matched, row, frequency, level_type
                )
                grouped_json.setdefault(key, {})[var_name] = cmor_entry
    if not exclude_extra_variables:
        _add_extra_variables(grouped_json, search_dirs, verbose)
    json_object = grouped_json.items()
    if not verbose:
        json_object = tqdm(
            json_object,
            total=len(grouped_json.items()),
            desc="💾 Writing JSON files",
            colour="green",
        )
    for (frequency, level_group), variable_entry in json_object:
        variable_entry = dict(sorted(variable_entry.items()))
        if "land" in level_group:
            level_group = level_group.replace("_land", "")
            filename = f"ERA5Land_{frequency}_{level_group}"
        else:
            filename = f"ERA5_{frequency}_{level_group}"

        cmor_json: dict[str, Any] = {
            "Header": {
                "Conventions": "CF-1.7 ODS-2.1",
                "approx_interval": f"{approx_interval_map[frequency]}",
                "cmor_version": "3.5",
                "data_specs_version": "2.1.0",
                "generic_levels": "",
                "int_missing_value": "-999",
                "mip_era": "CMIP6",
                "missing_value": "1e20",
                "product": "model-output",
                "realm": " ".join(
                    sorted(
                        {
                            v.get("modeling_realm", "")
                            for v in variable_entry.values()
                        }
                    )
                ).strip(),
                "table_date": f"{today_date}",
                "table_id": f"Table {filename}",
            },
            "variable_entry": variable_entry,
        }

        os.makedirs(json_output_path, exist_ok=True)
        output_path = os.path.join(json_output_path, f"{filename}.json")
        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(cmor_json, f, indent=4)

        if verbose:
            print(f"💾 Written file: {output_path}")


def harmonize_variables(
    json_dir: Path = json_output_path, verbose: bool = False
) -> None:
    """
    Post-process pass to fill in certain JSON values from higher-priority tables.

    This function iterates through all JSON files in the specified directory and
    updates the "variable_entry" and "Header" sections of each file.
    Currently it only populates empty "comment" fields cross-checking among all
    same variables of different frequencies.

    Parameters:
        json_dir (Path): Directory with generated CMOR JSON files.
        verbose (bool): Enable verbose output.

     Returns:
        None: The function modifies the JSON files in place.
    """
    grouped: dict[str, dict[str, dict]] = {}
    for json_file in sorted(json_dir.glob("ERA5*.json")):
        with open(json_file, "r", encoding="utf-8") as f:
            data = json.load(f)

        frequency = data["Header"]["frequency"] = (
            data["variable_entry"]
            .values()
            .__iter__()
            .__next__()
            .get("frequency")
        )
        variable_entry = data["variable_entry"]

        for var, content in variable_entry.items():
            grouped.setdefault(var, {}).setdefault(frequency, {})[
                "content"
            ] = content
            grouped[var][frequency]["file"] = json_file

    iterable = grouped.items()
    if not verbose:
        iterable = tqdm(
            iterable,
            total=len(grouped.items()),
            desc="✏️ updating JSON files",
            colour="yellow",
        )
    for var, freq_dict in iterable:
        preferred_comment = None
        for freq in ["mon", "day", "1hr"]:  # Priority order
            content = freq_dict.get(freq, {}).get("content", {})
            if (
                content.get("source_table", "").endswith(".json")
                and content.get("comment", "").strip()
            ):
                preferred_comment = content["comment"]
                break
        if not preferred_comment:
            continue

        for freq, info in freq_dict.items():
            content = info["content"]
            if not content.get("comment"):
                if verbose:
                    print(f"✏️ Updating comment → {var:15} | {freq:4}")
                content["comment"] = preferred_comment
                filepath = info["file"]
                with open(filepath, "r", encoding="utf-8") as f:
                    data = json.load(f)

                data["variable_entry"][var] = content
                with open(filepath, "w", encoding="utf-8") as f:
                    json.dump(data, f, indent=4)


if __name__ == "__main__":
    csv_to_cmor_json()
    harmonize_variables()
